#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gstglshader.h>

#define VALID_VARNAME_CHARS "abcdefghijklmnopqurstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"

#define trimleft(s,chars) while(s[0] && strchr(chars,s[0])) ++s;
#define trimright(s,chars) { \
	char *end; \
	end=s+strlen(s)-1; \
	while(end>=s && strchr(chars,end[0])) (end--)[0]=0; \
}

char *datatype[]={
	"bool",
	"int",
	"uint",
	"float",
	"vec2",
	"vec3",
	"vec4",
	"bvec2",
	"bvec3",
	"bvec4",
	"ivec2",
	"ivec3",
	"ivec4",
	"uvec2",
	"uvec3",
	"uvec4",
	"mat2",
	"mat3",
	"mat4",
	"mat2x2",
	"mat2x3",
	"mat2x4",
	"mat3x2",
	"mat3x3",
	"mat3x4",
	"mat4x2",
	"mat4x3",
	"mat4x4",
	0
};

typedef enum {
	_bool,
	_int,
	_uint,
	_float,
	_vec2,
	_vec3,
	_vec4,
	_bvec2,
	_bvec3,
	_bvec4,
	_ivec2,
	_ivec3,
	_ivec4,
	_uvec2,
	_uvec3,
	_uvec4,
	_mat2,
	_mat3,
	_mat4,
	_mat2x2,
	_mat2x3,
	_mat2x4,
	_mat3x2,
	_mat3x3,
	_mat3x4,
	_mat4x2,
	_mat4x3,
	_mat4x4,
	_datatypecount
} datatypeindex;

typedef struct variable_desc {
	datatypeindex	type;
	char		*name;
	int		arraysize;
	int		count;
	void		*value;

} variable_desc;

typedef struct variable_cache {
	char *source;
	struct variable_desc *var;
	unsigned long size;
	unsigned long count;

} variable_cache;

struct variable_cache *varcache;
unsigned long varcache_size;
unsigned long varcache_count;

char *vec_parsevalue(int n,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i;
	int j;
	int k;
	char *saveptr=value;
	char *saveptr2;
	char *t;
	char *u;

	ret->count=(arraysize)?arraysize*n:n;
	ret->value=g_malloc(sizeof(float)*ret->count);

	if (!arraysize) {
		t=strtok_r(value,",",&saveptr);

		for (i=0;i<ret->count;++i) {

			if (!t) return _saveptr+(saveptr-value);
		
			trimleft(t," \t");
			trimright(t," \t");
			if (!t[0]) return _saveptr+(saveptr-t);

			for (j=0; j<strlen(value); ++j) {
				if (!strchr("0123456789.-",value[j])) return _saveptr+(saveptr-t)+j;
			}

			((float*)ret->value)[i]=(float)atof(t);
			t=strtok_r(0,",",&saveptr);
		}

	} else {

		saveptr2=value;
		u=strtok_r(value,")",&saveptr2);

		for (k=0;k<arraysize;++k) {

			if (!u) return _saveptr+(saveptr2-value);

			trimleft(u," \t");
			trimright(u," \t");

			if (k) {
				if (u[0]!=',') return _saveptr+(u-value);
				++u;
				trimleft(u," \t");
			}
			
			if (strncmp(u,datatype[ret->type],strlen(datatype[ret->type]))) return _saveptr+(u-value);
		
			u+=strlen(datatype[ret->type]);
			trimleft(u," \t");
			if (u[0]!='(') return _saveptr+(u-value);
			++u;

			t=strtok_r(u,",",&saveptr);
			if (!t) return _saveptr+(u-value);

			for (i=0;i<n;++i) {

				trimleft(t," \t");
				trimright(t," \t");
				if (!t[0]) return _saveptr+(t-value);

				for (j=0; j<strlen(t); ++j) {
					if (!strchr("0123456789.-",t[j])) return _saveptr+(t-value)+j;
				}

				((float*)ret->value)[k*n+i]=(float)atof(t);
				t=strtok_r(0,",",&saveptr);
				if (i<(n-1) && !t) return _saveptr+(saveptr-value);
			
			}
			u=strtok_r(0,")",&saveptr2);
		}
	}
	return 0;
}

char *bvec_parsevalue(int n,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i;
	int k;
	char *saveptr=value;
	char *saveptr2;
	char *t;
	char *u;

	ret->count=(arraysize)?arraysize*n:n;
	ret->value=g_malloc(sizeof(char**)*ret->count);

	if (!arraysize) {
		t=strtok_r(value,",",&saveptr);

		for (i=0;i<ret->count;++i) {

			if (!t) return _saveptr+(saveptr-value);
		
			trimleft(t," \t");
			trimright(t," \t");
			if (!t[0]) return _saveptr+(saveptr-t);

			if (strcmp("true",value)||strcmp("false",value)) return _saveptr+(saveptr-t);

			((int*)ret->value)[i]=strcmp(t,"false");
			t=strtok_r(0,",",&saveptr);
		}

	} else {

		saveptr2=value;
		u=strtok_r(value,")",&saveptr2);

		for (k=0;k<arraysize;++k) {

			if (!u) return _saveptr+(saveptr2-value);

			trimleft(u," \t");
			trimright(u," \t");

			if (k) {
				if (u[0]!=',') return _saveptr+(u-value);
				++u;
				trimleft(u," \t");
			}
			
			if (strncmp(u,datatype[ret->type],strlen(datatype[ret->type]))) return _saveptr+(u-value);
		
			u+=strlen(datatype[ret->type]);
			trimleft(u," \t");
			if (u[0]!='(') return _saveptr+(u-value);
			++u;

			t=strtok_r(u,",",&saveptr);
			if (!t) return _saveptr+(u-value);

			for (i=0;i<n;++i) {

				trimleft(t," \t");
				trimright(t," \t");
				if (!t[0]) return _saveptr+(t-value);

				if (strcmp("true",t)||strcmp("false",t)) return _saveptr+(saveptr-t);

				((int*)ret->value)[k*n+i]=strcmp(t,"false");
				t=strtok_r(0,",",&saveptr);
				if (i<(n-1) && !t) return _saveptr+(saveptr-value);
			
			}
			u=strtok_r(0,")",&saveptr2);
		}
	}
	return 0;
}

char *ivec_parsevalue(int n,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i;
	int j;
	int k;
	char *saveptr=value;
	char *saveptr2;
	char *t;
	char *u;

	ret->count=(arraysize)?arraysize*n:n;
	ret->value=g_malloc(sizeof(int)*ret->count);

	if (!arraysize) {
		t=strtok_r(value,",",&saveptr);

		for (i=0;i<ret->count;++i) {

			if (!t) return _saveptr+(saveptr-value);
		
			trimleft(t," \t");
			trimright(t," \t");
			if (!t[0]) return _saveptr+(saveptr-t);

			for (j=0; j<strlen(value); ++j) {
				if (!strchr("0123456789-",value[j])) return _saveptr+(saveptr-t)+j;
			}

			((int*)ret->value)[i]=atoi(t);
			t=strtok_r(0,",",&saveptr);
		}

	} else {

		saveptr2=value;
		u=strtok_r(value,")",&saveptr2);

		for (k=0;k<arraysize;++k) {

			if (!u) return _saveptr+(saveptr2-value);

			trimleft(u," \t");
			trimright(u," \t");

			if (k) {
				if (u[0]!=',') return _saveptr+(u-value);
				++u;
				trimleft(u," \t");
			}
			
			if (strncmp(u,datatype[ret->type],strlen(datatype[ret->type]))) return _saveptr+(u-value);
		
			u+=strlen(datatype[ret->type]);
			trimleft(u," \t");
			if (u[0]!='(') return _saveptr+(u-value);
			++u;

			t=strtok_r(u,",",&saveptr);
			if (!t) return _saveptr+(u-value);

			for (i=0;i<n;++i) {

				trimleft(t," \t");
				trimright(t," \t");
				if (!t[0]) return _saveptr+(t-value);

				for (j=0; j<strlen(t); ++j) {
					if (!strchr("0123456789-",t[j])) return _saveptr+(t-value)+j;
				}

				((int*)ret->value)[k*n+i]=atoi(t);
				t=strtok_r(0,",",&saveptr);
				if (i<(n-1) && !t) return _saveptr+(saveptr-value);
			
			}
			u=strtok_r(0,")",&saveptr2);
		}
	}
	return 0;
}

char *uvec_parsevalue(int n,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i;
	int j;
	int k;
	char *saveptr=value;
	char *saveptr2;
	char *t;
	char *u;

	ret->count=(arraysize)?arraysize*n:n;
	ret->value=g_malloc(sizeof(unsigned int)*ret->count);

	if (!arraysize) {
		t=strtok_r(value,",",&saveptr);

		for (i=0;i<ret->count;++i) {

			if (!t) return _saveptr+(saveptr-value);
		
			trimleft(t," \t");
			trimright(t," \t");
			if (!t[0]) return _saveptr+(saveptr-t);

			for (j=0; j<strlen(value); ++j) {
				if (!strchr("0123456789",value[j])) return _saveptr+(saveptr-t)+j;
			}

			((unsigned int*)ret->value)[i]=atoi(t);
			t=strtok_r(0,",",&saveptr);
		}

	} else {

		saveptr2=value;
		u=strtok_r(value,")",&saveptr2);

		for (k=0;k<arraysize;++k) {

			if (!u) return _saveptr+(saveptr2-value);

			trimleft(u," \t");
			trimright(u," \t");

			if (k) {
				if (u[0]!=',') return _saveptr+(u-value);
				++u;
				trimleft(u," \t");
			}
			
			if (strncmp(u,datatype[ret->type],strlen(datatype[ret->type]))) return _saveptr+(u-value);
		
			u+=strlen(datatype[ret->type]);
			trimleft(u," \t");
			if (u[0]!='(') return _saveptr+(u-value);
			++u;

			t=strtok_r(u,",",&saveptr);
			if (!t) return _saveptr+(u-value);

			for (i=0;i<n;++i) {

				trimleft(t," \t");
				trimright(t," \t");
				if (!t[0]) return _saveptr+(t-value);

				for (j=0; j<strlen(t); ++j) {
					if (!strchr("0123456789",t[j])) return _saveptr+(t-value)+j;
				}

				((unsigned int*)ret->value)[k*n+i]=atoi(t);
				t=strtok_r(0,",",&saveptr);
				if (i<(n-1) && !t) return _saveptr+(saveptr-value);
			
			}
			u=strtok_r(0,")",&saveptr2);
		}
	}
	return 0;
}

char *mat_parsevalue(int n, int m,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i;
	int j;
	int k;
	char *saveptr=value;
	char *saveptr2;
	char *t;
	char *u;

	ret->count=(arraysize)?arraysize*n*m:n*m;
	ret->value=g_malloc(sizeof(float)*ret->count);

	if (!arraysize) {
		t=strtok_r(value,",",&saveptr);

		for (i=0;i<ret->count;++i) {

			if (!t) return _saveptr+(saveptr-value);
		
			trimleft(t," \t");
			trimright(t," \t");
			if (!t[0]) return _saveptr+(saveptr-t);

			for (j=0; j<strlen(value); ++j) {
				if (!strchr("0123456789.-",value[j])) return _saveptr+(saveptr-t)+j;
			}

			((float*)ret->value)[i]=(float)atof(t);
			t=strtok_r(0,",",&saveptr);
		}

	} else {

		saveptr2=value;
		u=strtok_r(value,")",&saveptr2);

		for (k=0;k<arraysize;++k) {

			if (!u) return _saveptr+(saveptr2-value);

			trimleft(u," \t");
			trimright(u," \t");

			if (k) {
				if (u[0]!=',') return _saveptr+(u-value);
				++u;
				trimleft(u," \t");
			}
			
			if (strncmp(u,datatype[ret->type],strlen(datatype[ret->type]))) return _saveptr+(u-value);
		
			u+=strlen(datatype[ret->type]);
			trimleft(u," \t");
			if (u[0]!='(') return _saveptr+(u-value);
			++u;

			t=strtok_r(u,",",&saveptr);
			if (!t) return _saveptr+(u-value);

			for (i=0;i<n*m;++i) {

				trimleft(t," \t");
				trimright(t," \t");
				if (!t[0]) return _saveptr+(t-value);

				for (j=0; j<strlen(t); ++j) {
					if (!strchr("0123456789.-",t[j])) return _saveptr+(t-value)+j;
				}

				((float*)ret->value)[k*n*m+i]=(float)atof(t);
				t=strtok_r(0,",",&saveptr);
				if (i<(n*m-1) && !t) return _saveptr+(saveptr-value);
			
			}
			u=strtok_r(0,")",&saveptr2);
		}
	}
	return 0;
}

char *parsevalue(char *vartype,char *varname,int arraysize,char *value,char *_saveptr,struct variable_desc *ret) {

	int i,j;
	char *t;
	char *saveptr=value;
	ret->name=varname;
	ret->arraysize=arraysize;

	switch(ret->type) {
		case _bool:
			ret->count=(arraysize)?arraysize:1;
			if (ret->count==1) {
				if (strcmp(value,"true") && strcmp(value,"false")) return _saveptr;
				ret->value=(void*)g_malloc(sizeof(int));
			 	((int*)ret->value)[0]=strcmp(value,"false");

			} else {
				ret->value=g_malloc(sizeof(int*)*ret->count);
				t=strtok_r(value,",",&saveptr);
				for (i=0;i<ret->count;++i) {
					if (!t) return _saveptr+(saveptr-value);
					trimleft(t," \t");
					trimright(t," \t");
					if (strcmp(t,"true") && strcmp(t,"false")) return _saveptr+(saveptr-t);
					((int*)ret->value)[i]=strcmp(t,"false");
					t=strtok_r(0,",",&saveptr);
				}
			}
			break;
		case _int:
			ret->count=(arraysize)?arraysize:1;
			if (ret->count==1) {
				for (j=0; j<strlen(value); ++j) {
					if (!strchr("-0123456789",value[j])) return _saveptr+j;
				}
				ret->value=(void*)g_malloc(sizeof(int));
				*((int*)ret->value)=atoi(value);
	
			} else {
				ret->value=g_malloc(sizeof(int)*ret->count);
				t=strtok_r(value,",",&saveptr);
	
				for (i=0;i<ret->count;++i) {

					if (!t) return _saveptr+(saveptr-value);
				
					trimleft(t," \t");
					trimright(t," \t");
					if (!t[0]) return _saveptr+(saveptr-t);

					for (j=0; j<strlen(value); ++j) {
						if (!strchr("-0123456789",value[j])) return _saveptr+(saveptr-t)+j;
					}

					((int*)ret->value)[i]=atoi(t);
					t=strtok_r(0,",",&saveptr);
				}
			}
			break;

		case _uint:
			ret->count=(arraysize)?arraysize:1;
			if (ret->count==1) {
				for (j=0; j<strlen(value); ++j) {
					if (!strchr("0123456789",value[j])) return _saveptr+j;
				}
				ret->value=(void*)g_malloc(sizeof(unsigned int));
				*((unsigned int*)ret->value)=atoi(value);
	
			} else {
				ret->value=g_malloc(sizeof(unsigned int)*ret->count);
				t=strtok_r(value,",",&saveptr);
	
				for (i=0;i<ret->count;++i) {

					if (!t) return _saveptr+(saveptr-value);
				
					trimleft(t," \t");
					trimright(t," \t");
					if (!t[0]) return _saveptr+(saveptr-t);

					for (j=0; j<strlen(value); ++j) {
						if (!strchr("0123456789",value[j])) return _saveptr+(saveptr-t)+j;
					}

					((unsigned int*)ret->value)[i]=atoi(t);
					t=strtok_r(0,",",&saveptr);
				}
			}
			break;

		case _float:
			ret->count=(arraysize)?arraysize:1;
			if (ret->count==1) {
				for (j=0; j<strlen(value); ++j) {
					if (!strchr("0123456789.-",value[j])) return _saveptr+j;
				}
				ret->value=(void*)g_malloc(sizeof(float));
				*((float*)ret->value)=(float)atof(value);
	
			} else {
				ret->value=g_malloc(sizeof(float)*ret->count);
				t=strtok_r(value,",",&saveptr);
	
				for (i=0;i<ret->count;++i) {

					if (!t) return _saveptr+(saveptr-value);
				
					trimleft(t," \t");
					trimright(t," \t");
					if (!t[0]) return _saveptr+(saveptr-t);

					for (j=0; j<strlen(value); ++j) {
						if (!strchr("0123456789.-",value[j])) return _saveptr+(saveptr-t)+j;
					}

					((float*)ret->value)[i]=(float)atof(t);
					t=strtok_r(0,",",&saveptr);
				}
			}
			break;

		case _vec2:
			return vec_parsevalue(2,arraysize,value,_saveptr,ret);
			break;

		case _bvec2:
			return bvec_parsevalue(2,arraysize,value,_saveptr,ret);
			break;

		case _ivec2:
			return ivec_parsevalue(2,arraysize,value,_saveptr,ret);
			break;

		case _uvec2:
			return uvec_parsevalue(2,arraysize,value,_saveptr,ret);
			break;

		case _vec3:
			return vec_parsevalue(3,arraysize,value,_saveptr,ret);
			break;

		case _bvec3:
			return bvec_parsevalue(3,arraysize,value,_saveptr,ret);
			break;

		case _ivec3:
			return uvec_parsevalue(3,arraysize,value,_saveptr,ret);
			break;

		case _uvec3:
			return uvec_parsevalue(3,arraysize,value,_saveptr,ret);
			break;

		case _vec4:
			return vec_parsevalue(4,arraysize,value,_saveptr,ret);
			break;

		case _bvec4:
			return bvec_parsevalue(4,arraysize,value,_saveptr,ret);
			break;

		case _ivec4:
			return ivec_parsevalue(4,arraysize,value,_saveptr,ret);
			break;

		case _uvec4:
			return uvec_parsevalue(4,arraysize,value,_saveptr,ret);
			break;

		case _mat2:
		case _mat2x2:
			return mat_parsevalue(2,2,arraysize,value,_saveptr,ret);
			break;

		case _mat2x3:
			return mat_parsevalue(2,3,arraysize,value,_saveptr,ret);
			break;

		case _mat3x2:
			return mat_parsevalue(3,2,arraysize,value,_saveptr,ret);
			break;

		case _mat2x4:
			return mat_parsevalue(2,4,arraysize,value,_saveptr,ret);
			break;

		case _mat4x2:
			return mat_parsevalue(4,2,arraysize,value,_saveptr,ret);
			break;
		
		case _mat3:
		case _mat3x3:
			return mat_parsevalue(3,3,arraysize,value,_saveptr,ret);
			break;

		case _mat3x4:
			return mat_parsevalue(3,4,arraysize,value,_saveptr,ret);
			break;

		case _mat4x3:
			return mat_parsevalue(4,3,arraysize,value,_saveptr,ret);
			break;

		case _mat4:
		case _mat4x4:
			return mat_parsevalue(4,4,arraysize,value,_saveptr,ret);
			break;

		default:
			break;
	}
	return 0;
}

int setuniformvariable(GstGLShader *shader,struct variable_desc *ret) {

	switch(ret->type) {
		case _bool:
			if (ret->arraysize) {
				gst_gl_shader_set_uniform_1iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_1i(shader,ret->name,((int*)ret->value)[0]);
			}
			break;
		case _int:
			if (ret->arraysize) {
				gst_gl_shader_set_uniform_1iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_1i(shader,ret->name,((int*)ret->value)[0]);
			}
			break;

		case _uint:
			if (ret->arraysize) {
				gst_gl_shader_set_uniform_1iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_1i(shader,ret->name,((unsigned int*)ret->value)[0]);
			}
			break;

		case _float:
			if (ret->arraysize) {
				gst_gl_shader_set_uniform_1fv(shader,ret->name,ret->count,(float*)ret->value);
			} else {
				gst_gl_shader_set_uniform_1f(shader,ret->name,((float*)ret->value)[0]);
			}
			break;

		case _vec2:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_2fv(shader,ret->name,ret->count,(float*)ret->value);
			} else {
				gst_gl_shader_set_uniform_2f(shader,ret->name,((float*)ret->value)[0],((float*)ret->value)[1]);
			}
			break;

		case _bvec2:
		case _ivec2:
		case _uvec2:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_2iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_2i(shader,ret->name,((int*)ret->value)[0],((int*)ret->value)[1]);
			}
			break;

		case _vec3:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_3fv(shader,ret->name,ret->count,(float*)ret->value);
			} else {
				gst_gl_shader_set_uniform_3f(shader,ret->name,((float*)ret->value)[0],((float*)ret->value)[1],((float*)ret->value)[2]);
			}
			break;

		case _bvec3:
		case _ivec3:
		case _uvec3:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_3iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_3i(shader,ret->name,((int*)ret->value)[0],((int*)ret->value)[1],((int*)ret->value)[2]);
			}
			break;

		case _vec4:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_4fv(shader,ret->name,ret->count,(float*)ret->value);
			} else {
				gst_gl_shader_set_uniform_4f(shader,ret->name,((float*)ret->value)[0],((float*)ret->value)[1],((float*)ret->value)[2],((float*)ret->value)[3]);
			}
			break;

		case _bvec4:
		case _ivec4:
		case _uvec4:
			if (ret->arraysize) { 
				gst_gl_shader_set_uniform_4iv(shader,ret->name,ret->count,(int*)ret->value);
			} else {
				gst_gl_shader_set_uniform_4i(shader,ret->name,((int*)ret->value)[0],((int*)ret->value)[1],((int*)ret->value)[2],((int*)ret->value)[3]);
			}
			break;

		case _mat2:
		case _mat2x2:
			gst_gl_shader_set_uniform_matrix_2fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat2x3:
			gst_gl_shader_set_uniform_matrix_2x3fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat3x2:
			gst_gl_shader_set_uniform_matrix_3x2fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat2x4:
			gst_gl_shader_set_uniform_matrix_2x4fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat4x2:
			gst_gl_shader_set_uniform_matrix_4x2fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;
		
		case _mat3:
		case _mat3x3:
			gst_gl_shader_set_uniform_matrix_3fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat3x4:
			gst_gl_shader_set_uniform_matrix_3x4fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat4x3:
			gst_gl_shader_set_uniform_matrix_4x3fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		case _mat4:
		case _mat4x4:
			gst_gl_shader_set_uniform_matrix_4fv(shader,ret->name,ret->count, 0,(float*)ret->value);
			break;

		default:
			return -1;
	}
	return 0;
}

char *parsename(char **varname,int *arraysize,char **saveptr) {
	char *t;
	char *i;
	int j;

	*arraysize=0;
	t=strtok_r(0,"=",saveptr);
	if (!t) return *saveptr;

	trimleft(t," \t");
	trimright(t," \t");

	i=strchr(t,'[');
	if (!i) {
		if (!t[0]) return t;
        	for(j=0;j<strlen(t);++j) {
			if (!strchr(VALID_VARNAME_CHARS,t[j])) return t+j;
		}
		*varname=g_strdup(t);
	} else {

		char *i2;
		char *c;

		i2=strchr(i+1,']');
		if (!i2) return i+1;
		*i=0;

		if (!t[0]) return t;
        	for(j=0;j<strlen(t);++j) {
			if (!strchr(VALID_VARNAME_CHARS,t[j])) return t;
		}

		*varname=g_strdup(t);
		*i='[';

		for(c=i+1;c<i2;++c) if (*c<'0'||*c>'9') return c;
		
		*i2=0;
		*arraysize=atoi(i+1);
		*i2=']';

		if (!*arraysize) return i+1;
	}

	return 0;
}

void gst_gl_shadervariables_cache_delete(char *variables) {

	int i;
	struct variable_cache *cache;

	for(i=0;i<varcache_count;++i) {
		cache=&varcache[i];
		if (cache->source==variables) {
			int j;
			for(j=0;j<cache->count;++j) {
				g_free(cache->var[j].name);
				g_free(cache->var[j].value);
			}
			cache->count=0;
			cache->source=0;
			return;
		}
	}
}

int gst_gl_shadervariables_parse(GstGLShader *shader,char *variables) {
	char *p=0;
	char *p0;
	char *e;
	char e1=0;
	char *t=0;
	char *varname;
	char *vartype;
	char *varvalue;
	int arraysize=0;
	char *saveptr=variables;
	int line=1;
	char *lim;
	int i;
	int len;
	struct variable_cache *cache;
	int freeone=-1;

	if (!variables) return 0;

	if (!varcache) {
		varcache_size=64;
		varcache=g_malloc(sizeof(struct variable_cache)*varcache_size);
		if (!varcache) {
			GST_ERROR("g_malloc failed !");
			exit(1);
		}
		varcache_count=0;

	} else {
		for(i=0;i<varcache_count;++i) {
			cache=&varcache[i];
			if (cache->source==variables) {
				int j;
				for(j=0;j<cache->count;++j) {
					setuniformvariable(shader,&cache->var[j]);
				}
				return 0;
			} else {
				if (!cache->source && freeone<0) {
				       	freeone=i;
				}
			}

		}

		if (varcache_size==varcache_count && freeone<0) {
			varcache_size+=64;
			varcache=g_realloc(varcache,varcache_size);
			if (!varcache) {
				GST_ERROR("g_realloc failed !");
				exit(1);
			}
		}
	}

	if (freeone<0) freeone=varcache_count++;
	cache=&varcache[freeone];
	cache->source=variables;
	cache->var=0;
	cache->size=0;
	cache->count=0;

	p0=variables;
	lim=variables+strlen(variables);
	e=strchr(p0,';');
	while(p0<lim) {

		if (!e) goto parse_error;

		e1=e[1];
		e[1]=0;
		p=g_strdup(p0);
		e[1]=e1;

		trimright(p," \t");
		trimleft(p," \t\n");

		t=strtok_r(p," \t",&saveptr);
		if (!t) goto parse_error;
		trimleft(t," \t");
		trimright(t," \t\n");

		if (t[0]) {

			if (!cache->var) {
				cache->size=64;
				cache->var=g_malloc(sizeof(struct variable_desc)*cache->size);
				cache->count=0;
			} else {
				if (cache->size==cache->count) {
					cache->size+=64;
					cache->var=g_realloc(cache->var,sizeof(struct variable_desc)*cache->size);
					if (!cache->var) {
						GST_ERROR("g_realloc failed !");
						exit(1);
					}
				}
			}

			if (!strcmp(t,"const")) {
				t=strtok_r(0," \t",&saveptr);
				if (!t) goto parse_error;
				trimleft(t," \t");
				if (!t[0]) goto parse_error;
			}

			for(i=0;i<_datatypecount;++i) {
				if (!strcmp(t,datatype[i])) {
					cache->var[cache->count].type=(datatypeindex)i;
					break;
				}
			}
			if (i==_datatypecount) goto parse_error;

			vartype=g_strdup(t);
			GST_INFO("vartype : '%s'\n",vartype);

			trimleft(saveptr," \t");
			t=saveptr;
			if (*saveptr=='=') goto parse_error;

			t=parsename(&varname,&arraysize,&saveptr);
			if (t) goto parse_error;

			trimright(varname," \t");
			GST_INFO("varname : '%s'\n",varname);
			GST_INFO("arraysize : %d\n",arraysize);

			t=strtok_r(0,"(",&saveptr);
			if (!t) goto parse_error;
			trimleft(t," \t");
			trimright(t," \t");

			if (arraysize) {
				char *s=g_malloc(strlen(vartype)+32);
				sprintf(s,"%s[%d]",vartype,arraysize);
				if (strcmp(t,s)) goto parse_error;
			} else {
				if (strcmp(t,vartype)) goto parse_error;
			}

			t=strtok_r(0,";",&saveptr);
			if (!t) goto parse_error;
			trimleft(t," \t");
			trimright(t," \t");

			if (!t[0]) goto parse_error;
			if (*(saveptr-2)!=')') goto parse_error;
			*(saveptr-2)=0;
			if (!t[0]) goto parse_error;

			varvalue=g_strdup(t);
			GST_INFO("value: %s\n\n",varvalue);

			t=saveptr;
			if (t[0]) goto parse_error;

			len=strlen(varvalue);
			t=parsevalue(vartype,varname,arraysize,varvalue,saveptr,&cache->var[cache->count]);
			if (t) {
				t-=len;
				goto parse_error;
			}
			setuniformvariable(shader,&cache->var[cache->count]);
			++cache->count;

			fflush(0);
		}

		// im tired... tell me where is the bug..
		//g_free(p);
		p0=e+1;
		++line;
		e=strchr(p0,';');
	}

	return 0;

parse_error:
	if (!t) {
		t=saveptr;
	}
	if (!e) {
		t=p=p0;
	} else {
		e[1]=0;
		trimleft(p0," \t\n");
		GST_ERROR("%s\n",p0);
		e[1]=e1;
		for (i=0; i<t-p; ++i) fprintf(stderr," ");
		fprintf(stderr,"^\n");
	}
	GST_ERROR("parse error on line %d, position %d (%s)",line,t-p,t);
	return -1;

}


